"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings
from . import cortex_axon_agent_pb2 as cortex__axon__agent__pb2
GRPC_GENERATED_VERSION = '1.68.1'
GRPC_VERSION = grpc.__version__
_version_not_supported = False
try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True
if _version_not_supported:
    raise RuntimeError(f'The grpc package installed is at version {GRPC_VERSION},' + f' but the generated code in cortex_axon_agent_pb2_grpc.py depends on' + f' grpcio>={GRPC_GENERATED_VERSION}.' + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}' + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.')

class AxonAgentStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.RegisterHandler = channel.unary_unary('/cortex.axon.AxonAgent/RegisterHandler', request_serializer=cortex__axon__agent__pb2.RegisterHandlerRequest.SerializeToString, response_deserializer=cortex__axon__agent__pb2.RegisterHandlerResponse.FromString, _registered_method=True)
        self.UnregisterHandler = channel.unary_unary('/cortex.axon.AxonAgent/UnregisterHandler', request_serializer=cortex__axon__agent__pb2.UnregisterHandlerRequest.SerializeToString, response_deserializer=cortex__axon__agent__pb2.UnregisterHandlerResponse.FromString, _registered_method=True)
        self.ListHandlers = channel.unary_unary('/cortex.axon.AxonAgent/ListHandlers', request_serializer=cortex__axon__agent__pb2.ListHandlersRequest.SerializeToString, response_deserializer=cortex__axon__agent__pb2.ListHandlersResponse.FromString, _registered_method=True)
        self.GetHandlerHistory = channel.unary_unary('/cortex.axon.AxonAgent/GetHandlerHistory', request_serializer=cortex__axon__agent__pb2.GetHandlerHistoryRequest.SerializeToString, response_deserializer=cortex__axon__agent__pb2.GetHandlerHistoryResponse.FromString, _registered_method=True)
        self.Dispatch = channel.stream_stream('/cortex.axon.AxonAgent/Dispatch', request_serializer=cortex__axon__agent__pb2.DispatchRequest.SerializeToString, response_deserializer=cortex__axon__agent__pb2.DispatchMessage.FromString, _registered_method=True)
        self.ReportInvocation = channel.unary_unary('/cortex.axon.AxonAgent/ReportInvocation', request_serializer=cortex__axon__agent__pb2.ReportInvocationRequest.SerializeToString, response_deserializer=cortex__axon__agent__pb2.ReportInvocationResponse.FromString, _registered_method=True)

class AxonAgentServicer(object):
    """Missing associated documentation comment in .proto file."""

    def RegisterHandler(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UnregisterHandler(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListHandlers(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetHandlerHistory(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Dispatch(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ReportInvocation(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

def add_AxonAgentServicer_to_server(servicer, server):
    rpc_method_handlers = {'RegisterHandler': grpc.unary_unary_rpc_method_handler(servicer.RegisterHandler, request_deserializer=cortex__axon__agent__pb2.RegisterHandlerRequest.FromString, response_serializer=cortex__axon__agent__pb2.RegisterHandlerResponse.SerializeToString), 'UnregisterHandler': grpc.unary_unary_rpc_method_handler(servicer.UnregisterHandler, request_deserializer=cortex__axon__agent__pb2.UnregisterHandlerRequest.FromString, response_serializer=cortex__axon__agent__pb2.UnregisterHandlerResponse.SerializeToString), 'ListHandlers': grpc.unary_unary_rpc_method_handler(servicer.ListHandlers, request_deserializer=cortex__axon__agent__pb2.ListHandlersRequest.FromString, response_serializer=cortex__axon__agent__pb2.ListHandlersResponse.SerializeToString), 'GetHandlerHistory': grpc.unary_unary_rpc_method_handler(servicer.GetHandlerHistory, request_deserializer=cortex__axon__agent__pb2.GetHandlerHistoryRequest.FromString, response_serializer=cortex__axon__agent__pb2.GetHandlerHistoryResponse.SerializeToString), 'Dispatch': grpc.stream_stream_rpc_method_handler(servicer.Dispatch, request_deserializer=cortex__axon__agent__pb2.DispatchRequest.FromString, response_serializer=cortex__axon__agent__pb2.DispatchMessage.SerializeToString), 'ReportInvocation': grpc.unary_unary_rpc_method_handler(servicer.ReportInvocation, request_deserializer=cortex__axon__agent__pb2.ReportInvocationRequest.FromString, response_serializer=cortex__axon__agent__pb2.ReportInvocationResponse.SerializeToString)}
    generic_handler = grpc.method_handlers_generic_handler('cortex.axon.AxonAgent', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('cortex.axon.AxonAgent', rpc_method_handlers)

class AxonAgent(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def RegisterHandler(request, target, options=(), channel_credentials=None, call_credentials=None, insecure=False, compression=None, wait_for_ready=None, timeout=None, metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cortex.axon.AxonAgent/RegisterHandler', cortex__axon__agent__pb2.RegisterHandlerRequest.SerializeToString, cortex__axon__agent__pb2.RegisterHandlerResponse.FromString, options, channel_credentials, insecure, call_credentials, compression, wait_for_ready, timeout, metadata, _registered_method=True)

    @staticmethod
    def UnregisterHandler(request, target, options=(), channel_credentials=None, call_credentials=None, insecure=False, compression=None, wait_for_ready=None, timeout=None, metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cortex.axon.AxonAgent/UnregisterHandler', cortex__axon__agent__pb2.UnregisterHandlerRequest.SerializeToString, cortex__axon__agent__pb2.UnregisterHandlerResponse.FromString, options, channel_credentials, insecure, call_credentials, compression, wait_for_ready, timeout, metadata, _registered_method=True)

    @staticmethod
    def ListHandlers(request, target, options=(), channel_credentials=None, call_credentials=None, insecure=False, compression=None, wait_for_ready=None, timeout=None, metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cortex.axon.AxonAgent/ListHandlers', cortex__axon__agent__pb2.ListHandlersRequest.SerializeToString, cortex__axon__agent__pb2.ListHandlersResponse.FromString, options, channel_credentials, insecure, call_credentials, compression, wait_for_ready, timeout, metadata, _registered_method=True)

    @staticmethod
    def GetHandlerHistory(request, target, options=(), channel_credentials=None, call_credentials=None, insecure=False, compression=None, wait_for_ready=None, timeout=None, metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cortex.axon.AxonAgent/GetHandlerHistory', cortex__axon__agent__pb2.GetHandlerHistoryRequest.SerializeToString, cortex__axon__agent__pb2.GetHandlerHistoryResponse.FromString, options, channel_credentials, insecure, call_credentials, compression, wait_for_ready, timeout, metadata, _registered_method=True)

    @staticmethod
    def Dispatch(request_iterator, target, options=(), channel_credentials=None, call_credentials=None, insecure=False, compression=None, wait_for_ready=None, timeout=None, metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/cortex.axon.AxonAgent/Dispatch', cortex__axon__agent__pb2.DispatchRequest.SerializeToString, cortex__axon__agent__pb2.DispatchMessage.FromString, options, channel_credentials, insecure, call_credentials, compression, wait_for_ready, timeout, metadata, _registered_method=True)

    @staticmethod
    def ReportInvocation(request, target, options=(), channel_credentials=None, call_credentials=None, insecure=False, compression=None, wait_for_ready=None, timeout=None, metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cortex.axon.AxonAgent/ReportInvocation', cortex__axon__agent__pb2.ReportInvocationRequest.SerializeToString, cortex__axon__agent__pb2.ReportInvocationResponse.FromString, options, channel_credentials, insecure, call_credentials, compression, wait_for_ready, timeout, metadata, _registered_method=True)


import functools
import inspect
from abc import abstractmethod

from generated import (
    cortex_axon_agent_pb2,
)


class CortexAnnotation:
    def __init__(self, timeout_ms: int = 0):
        self.func = None
        self.name = None
        self.timeout = timeout_ms

    def set_func(self, func):
        if self.func and self.func != func:
            raise ValueError("Cannot set func twice")
        self.name = func.__name__
        self.func = func

    def __call__(self, *args, **kwargs):
        if not self.func:
            raise ValueError("Handler not initialized")
        return self.func(*args, **kwargs)

    def __get__(self, instance, owner):
        return self.func.__get__(instance, owner)

    @abstractmethod
    def get_handler_options(self) -> list[cortex_axon_agent_pb2.HandlerOption]:
        pass


class CortexScheduled(CortexAnnotation):
    def __init__(
        self,
        # interval is a string that represents a time duration, eg "5s" or "3h"
        interval: str = None,
        cron: str = None,  # cron is a string that represents a cron expression, eg "* * * * *"
        # run_now is a boolean that determines if the handler should be run immediately
        run_now: bool = True,
        # timeout_ms is an integer that represents the timeout in milliseconds
        timeout_ms: int = 0,
    ):
        super().__init__(timeout_ms)
        self.interval = interval
        self.cron = cron
        self.run_now = run_now

    def get_handler_options(self) -> list[cortex_axon_agent_pb2.HandlerOption]:
        options = []
        if self.interval:
            options.append(
                cortex_axon_agent_pb2.HandlerOption(
                    invoke=cortex_axon_agent_pb2.HandlerInvokeOption(
                        type=cortex_axon_agent_pb2.RUN_INTERVAL, value=self.interval
                    )
                )
            )
        if self.cron:
            options.append(
                cortex_axon_agent_pb2.HandlerOption(
                    invoke=cortex_axon_agent_pb2.HandlerInvokeOption(
                        type=cortex_axon_agent_pb2.CRON_SCHEDULE, value=self.cron
                    )
                )
            )
        if self.run_now:
            options.append(
                cortex_axon_agent_pb2.HandlerOption(
                    invoke=cortex_axon_agent_pb2.HandlerInvokeOption(
                        type=cortex_axon_agent_pb2.RUN_NOW,
                    )
                )
            )
        return options


class CortexWebhook(CortexAnnotation):
    def __init__(self, id: str, timeout_ms: int = 0):
        super().__init__(timeout_ms)
        self.id = id

    def get_handler_options(self) -> list[cortex_axon_agent_pb2.HandlerOption]:
        return [
            cortex_axon_agent_pb2.HandlerOption(
                invoke=cortex_axon_agent_pb2.HandlerInvokeOption(
                    type=cortex_axon_agent_pb2.WEBHOOK, value=self.id
                )
            )
        ]


class CortexHandler(CortexAnnotation):
    def __init__(self, timeout_ms: int = 0):
        super().__init__(timeout_ms)

    def get_handler_options(self) -> list[cortex_axon_agent_pb2.HandlerOption]:
        return [
            cortex_axon_agent_pb2.HandlerOption(
                invoke=cortex_axon_agent_pb2.HandlerInvokeOption(
                    type=cortex_axon_agent_pb2.INVOKE
                )
            )
        ]


def cortex_handler(*a, **k):
    handler = CortexHandler()
    name = cortex_handler.__name__

    def real_decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            retval = f(*args, **kwargs)
            return retval

        existing = getattr(f, name, None)
        if existing:
            raise ValueError(f'Attribute type ${name} already exists')
        setattr(wrapper, name, handler)
        return wrapper

    return real_decorator


def cortex_scheduled(*a, **k):

    interval = k.get("interval")
    cron = k.get("cron")
    run_now = k.get("run_now", True)
    name = cortex_scheduled.__name__

    handler = CortexScheduled(
        interval=interval, cron=cron, run_now=run_now
    )

    def real_decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            retval = f(*args, **kwargs)
            return retval

        existing = getattr(f, name, None)
        if existing:
            raise ValueError(f'Attribute type ${name} already exists')
        setattr(wrapper, name, handler)
        return wrapper

    return real_decorator


def cortex_webhook(*a, **k):
    id = k.get("id")
    handler = CortexWebhook(id=id)
    name = cortex_webhook.__name__

    def real_decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            retval = f(*args, **kwargs)
            return retval

        existing = getattr(f, name, None)
        if existing:
            raise ValueError(f'Attribute type ${name} already exists')
        setattr(wrapper, name, handler)
        return wrapper

    return real_decorator

handler_types =[cortex_scheduled, cortex_webhook, cortex_handler] 

def find_annotated_methods(scope: dict = None):
    handlers = []
    for _, func in scope.items():
        if not inspect.isfunction(func):
            continue
    
        for type in handler_types:
            h = getattr(func, type.__name__, None)
            if isinstance(h, CortexAnnotation):
                h.set_func(func)
                handlers.append(h)
                break
        
    return handlers



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(_runtime_version.Domain.PUBLIC, 5, 28, 1, '', 'cortex-axon-agent.proto')
_sym_db = _symbol_database.Default()
from . import common_pb2 as common__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17cortex-axon-agent.proto\x12\x0bcortex.axon\x1a\x0ccommon.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\x84\x01\n\x16RegisterHandlerRequest\x12\x13\n\x0bdispatch_id\x18\x01 \x01(\t\x12\x14\n\x0chandler_name\x18\x02 \x01(\t\x12\x12\n\ntimeout_ms\x18\x03 \x01(\x05\x12+\n\x07options\x18\x04 \x03(\x0b2\x1a.cortex.axon.HandlerOption"R\n\x13HandlerInvokeOption\x12,\n\x04type\x18\x01 \x01(\x0e2\x1e.cortex.axon.HandlerInvokeType\x12\r\n\x05value\x18\x02 \x01(\t"M\n\rHandlerOption\x122\n\x06invoke\x18\x01 \x01(\x0b2 .cortex.axon.HandlerInvokeOptionH\x00B\x08\n\x06option"H\n\x17RegisterHandlerResponse\x12!\n\x05error\x18\x01 \x01(\x0b2\x12.cortex.axon.Error\x12\n\n\x02id\x18\x02 \x01(\t"&\n\x18UnregisterHandlerRequest\x12\n\n\x02id\x18\x01 \x01(\t">\n\x19UnregisterHandlerResponse\x12!\n\x05error\x18\x01 \x01(\x0b2\x12.cortex.axon.Error"\x15\n\x13ListHandlersRequest"\xbf\x01\n\x0bHandlerInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12+\n\x07options\x18\x02 \x03(\x0b2\x1a.cortex.axon.HandlerOption\x12\x13\n\x0bdispatch_id\x18\x14 \x01(\t\x12\n\n\x02id\x18\x15 \x01(\t\x12A\n\x1dlast_invoked_client_timestamp\x18d \x01(\x0b2\x1a.google.protobuf.Timestamp\x12\x11\n\tis_active\x18e \x01(\x08"e\n\x14ListHandlersResponse\x12!\n\x05error\x18\x01 \x01(\x0b2\x12.cortex.axon.Error\x12*\n\x08handlers\x18\x02 \x03(\x0b2\x18.cortex.axon.HandlerInfo">\n\x0fDispatchRequest\x12\x13\n\x0bdispatch_id\x18\x01 \x01(\t\x12\x16\n\x0eclient_version\x18\x02 \x01(\t"\x82\x01\n\x0fDispatchMessage\x12.\n\x04type\x18\x01 \x01(\x0e2 .cortex.axon.DispatchMessageType\x124\n\x06invoke\x18\n \x01(\x0b2".cortex.axon.DispatchHandlerInvokeH\x00B\t\n\x07message"\x9a\x02\n\x15DispatchHandlerInvoke\x12\x15\n\rinvocation_id\x18\x01 \x01(\t\x12\x13\n\x0bdispatch_id\x18\x02 \x01(\t\x12\x12\n\nhandler_id\x18\x03 \x01(\t\x12\x14\n\x0chandler_name\x18\x04 \x01(\t\x12\x12\n\ntimeout_ms\x18\n \x01(\x05\x12.\n\x06reason\x18\x0b \x01(\x0e2\x1e.cortex.axon.HandlerInvokeType\x12:\n\x04args\x18\x14 \x03(\x0b2,.cortex.axon.DispatchHandlerInvoke.ArgsEntry\x1a+\n\tArgsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x028\x01"T\n\x03Log\x12\r\n\x05level\x18\x01 \x01(\t\x12-\n\ttimestamp\x18\x02 \x01(\x0b2\x1a.google.protobuf.Timestamp\x12\x0f\n\x07message\x18\x03 \x01(\t"\xa6\x02\n\x17ReportInvocationRequest\x12:\n\x0ehandler_invoke\x18\x01 \x01(\x0b2".cortex.axon.DispatchHandlerInvoke\x12:\n\x16start_client_timestamp\x18d \x01(\x0b2\x1a.google.protobuf.Timestamp\x12\x13\n\x0bduration_ms\x18e \x01(\x05\x12,\n\x06result\x18\xc8\x01 \x01(\x0b2\x19.cortex.axon.InvokeResultH\x00\x12$\n\x05error\x18\xc9\x01 \x01(\x0b2\x12.cortex.axon.ErrorH\x00\x12\x1f\n\x04logs\x18\xac\x02 \x03(\x0b2\x10.cortex.axon.LogB\t\n\x07message"=\n\x18ReportInvocationResponse\x12!\n\x05error\x18\x01 \x01(\x0b2\x12.cortex.axon.Error"\x1d\n\x0cInvokeResult\x12\r\n\x05value\x18\n \x01(\t"\xb2\x01\n\x18GetHandlerHistoryRequest\x12\x14\n\x0chandler_name\x18\x01 \x01(\t\x12.\n\nstart_time\x18\x02 \x01(\x0b2\x1a.google.protobuf.Timestamp\x12,\n\x08end_time\x18\x03 \x01(\x0b2\x1a.google.protobuf.Timestamp\x12\x14\n\x0cinclude_logs\x18\x04 \x01(\x08\x12\x0c\n\x04tail\x18\x05 \x01(\x05"\xf8\x02\n\x10HandlerExecution\x12\x14\n\x0chandler_name\x18\x01 \x01(\t\x12\x12\n\nhandler_id\x18\x02 \x01(\t\x12\x15\n\rinvocation_id\x18\x03 \x01(\t\x12\x13\n\x0bdispatch_id\x18\x04 \x01(\t\x12<\n\x18publish_server_timestamp\x18\n \x01(\x0b2\x1a.google.protobuf.Timestamp\x12<\n\x18receive_server_timestamp\x18\x0b \x01(\x0b2\x1a.google.protobuf.Timestamp\x12:\n\x16start_client_timestamp\x18\x0c \x01(\x0b2\x1a.google.protobuf.Timestamp\x12\x13\n\x0bduration_ms\x18\r \x01(\x05\x12!\n\x05error\x18\x14 \x01(\x0b2\x12.cortex.axon.Error\x12\x1e\n\x04logs\x18\x1e \x03(\x0b2\x10.cortex.axon.Log"n\n\x19GetHandlerHistoryResponse\x12!\n\x05error\x18\x01 \x01(\x0b2\x12.cortex.axon.Error\x12.\n\x07history\x18\x02 \x03(\x0b2\x1d.cortex.axon.HandlerExecution*^\n\x11HandlerInvokeType\x12\n\n\x06INVOKE\x10\x00\x12\x0b\n\x07RUN_NOW\x10\x01\x12\x11\n\rCRON_SCHEDULE\x10\x02\x12\x10\n\x0cRUN_INTERVAL\x10\x03\x12\x0b\n\x07WEBHOOK\x10\x04*w\n\x13DispatchMessageType\x12\x1e\n\x1aDISPATCH_MESSAGE_TYPE_NONE\x10\x00\x12\x1b\n\x17DISPATCH_MESSAGE_INVOKE\x10\x01\x12#\n\x1fDISPATCH_MESSAGE_WORK_COMPLETED\x10\x022\xb3\x04\n\tAxonAgent\x12\\\n\x0fRegisterHandler\x12#.cortex.axon.RegisterHandlerRequest\x1a$.cortex.axon.RegisterHandlerResponse\x12b\n\x11UnregisterHandler\x12%.cortex.axon.UnregisterHandlerRequest\x1a&.cortex.axon.UnregisterHandlerResponse\x12S\n\x0cListHandlers\x12 .cortex.axon.ListHandlersRequest\x1a!.cortex.axon.ListHandlersResponse\x12b\n\x11GetHandlerHistory\x12%.cortex.axon.GetHandlerHistoryRequest\x1a&.cortex.axon.GetHandlerHistoryResponse\x12J\n\x08Dispatch\x12\x1c.cortex.axon.DispatchRequest\x1a\x1c.cortex.axon.DispatchMessage(\x010\x01\x12_\n\x10ReportInvocation\x12$.cortex.axon.ReportInvocationRequest\x1a%.cortex.axon.ReportInvocationResponseB\x1cZ\x1agithub.com/cortexapps/axonb\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'cortex_axon_agent_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    _globals['DESCRIPTOR']._loaded_options = None
    _globals['DESCRIPTOR']._serialized_options = b'Z\x1agithub.com/cortexapps/axon'
    _globals['_DISPATCHHANDLERINVOKE_ARGSENTRY']._loaded_options = None
    _globals['_DISPATCHHANDLERINVOKE_ARGSENTRY']._serialized_options = b'8\x01'
    _globals['_HANDLERINVOKETYPE']._serialized_start = 2514
    _globals['_HANDLERINVOKETYPE']._serialized_end = 2608
    _globals['_DISPATCHMESSAGETYPE']._serialized_start = 2610
    _globals['_DISPATCHMESSAGETYPE']._serialized_end = 2729
    _globals['_REGISTERHANDLERREQUEST']._serialized_start = 88
    _globals['_REGISTERHANDLERREQUEST']._serialized_end = 220
    _globals['_HANDLERINVOKEOPTION']._serialized_start = 222
    _globals['_HANDLERINVOKEOPTION']._serialized_end = 304
    _globals['_HANDLEROPTION']._serialized_start = 306
    _globals['_HANDLEROPTION']._serialized_end = 383
    _globals['_REGISTERHANDLERRESPONSE']._serialized_start = 385
    _globals['_REGISTERHANDLERRESPONSE']._serialized_end = 457
    _globals['_UNREGISTERHANDLERREQUEST']._serialized_start = 459
    _globals['_UNREGISTERHANDLERREQUEST']._serialized_end = 497
    _globals['_UNREGISTERHANDLERRESPONSE']._serialized_start = 499
    _globals['_UNREGISTERHANDLERRESPONSE']._serialized_end = 561
    _globals['_LISTHANDLERSREQUEST']._serialized_start = 563
    _globals['_LISTHANDLERSREQUEST']._serialized_end = 584
    _globals['_HANDLERINFO']._serialized_start = 587
    _globals['_HANDLERINFO']._serialized_end = 778
    _globals['_LISTHANDLERSRESPONSE']._serialized_start = 780
    _globals['_LISTHANDLERSRESPONSE']._serialized_end = 881
    _globals['_DISPATCHREQUEST']._serialized_start = 883
    _globals['_DISPATCHREQUEST']._serialized_end = 945
    _globals['_DISPATCHMESSAGE']._serialized_start = 948
    _globals['_DISPATCHMESSAGE']._serialized_end = 1078
    _globals['_DISPATCHHANDLERINVOKE']._serialized_start = 1081
    _globals['_DISPATCHHANDLERINVOKE']._serialized_end = 1363
    _globals['_DISPATCHHANDLERINVOKE_ARGSENTRY']._serialized_start = 1320
    _globals['_DISPATCHHANDLERINVOKE_ARGSENTRY']._serialized_end = 1363
    _globals['_LOG']._serialized_start = 1365
    _globals['_LOG']._serialized_end = 1449
    _globals['_REPORTINVOCATIONREQUEST']._serialized_start = 1452
    _globals['_REPORTINVOCATIONREQUEST']._serialized_end = 1746
    _globals['_REPORTINVOCATIONRESPONSE']._serialized_start = 1748
    _globals['_REPORTINVOCATIONRESPONSE']._serialized_end = 1809
    _globals['_INVOKERESULT']._serialized_start = 1811
    _globals['_INVOKERESULT']._serialized_end = 1840
    _globals['_GETHANDLERHISTORYREQUEST']._serialized_start = 1843
    _globals['_GETHANDLERHISTORYREQUEST']._serialized_end = 2021
    _globals['_HANDLEREXECUTION']._serialized_start = 2024
    _globals['_HANDLEREXECUTION']._serialized_end = 2400
    _globals['_GETHANDLERHISTORYRESPONSE']._serialized_start = 2402
    _globals['_GETHANDLERHISTORYRESPONSE']._serialized_end = 2512
    _globals['_AXONAGENT']._serialized_start = 2732
    _globals['_AXONAGENT']._serialized_end = 3295


from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional
DESCRIPTOR: _descriptor.FileDescriptor

class CallRequest(_message.Message):
    __slots__ = ('method', 'path', 'content_type', 'body')
    METHOD_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    CONTENT_TYPE_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    method: str
    path: str
    content_type: str
    body: str

    def __init__(self, method: _Optional[str]=..., path: _Optional[str]=..., content_type: _Optional[str]=..., body: _Optional[str]=...) -> None:
        ...

class CallResponse(_message.Message):
    __slots__ = ('status_code', 'status', 'headers', 'body')

    class HeadersEntry(_message.Message):
        __slots__ = ('key', 'value')
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str

        def __init__(self, key: _Optional[str]=..., value: _Optional[str]=...) -> None:
            ...
    STATUS_CODE_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    HEADERS_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    status_code: int
    status: str
    headers: _containers.ScalarMap[str, str]
    body: str

    def __init__(self, status_code: _Optional[int]=..., status: _Optional[str]=..., headers: _Optional[_Mapping[str, str]]=..., body: _Optional[str]=...) -> None:
        ...


"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(_runtime_version.Domain.PUBLIC, 5, 28, 1, '', 'common.proto')
_sym_db = _symbol_database.Default()
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0ccommon.proto\x12\x0bcortex.axon"&\n\x05Error\x12\x0c\n\x04code\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\tB\x1cZ\x1agithub.com/cortexapps/axonb\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'common_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    _globals['DESCRIPTOR']._loaded_options = None
    _globals['DESCRIPTOR']._serialized_options = b'Z\x1agithub.com/cortexapps/axon'
    _globals['_ERROR']._serialized_start = 29
    _globals['_ERROR']._serialized_end = 67


import common_pb2 as _common_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union
DESCRIPTOR: _descriptor.FileDescriptor

class HandlerInvokeType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    INVOKE: _ClassVar[HandlerInvokeType]
    RUN_NOW: _ClassVar[HandlerInvokeType]
    CRON_SCHEDULE: _ClassVar[HandlerInvokeType]
    RUN_INTERVAL: _ClassVar[HandlerInvokeType]
    WEBHOOK: _ClassVar[HandlerInvokeType]

class DispatchMessageType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    DISPATCH_MESSAGE_TYPE_NONE: _ClassVar[DispatchMessageType]
    DISPATCH_MESSAGE_INVOKE: _ClassVar[DispatchMessageType]
    DISPATCH_MESSAGE_WORK_COMPLETED: _ClassVar[DispatchMessageType]
INVOKE: HandlerInvokeType
RUN_NOW: HandlerInvokeType
CRON_SCHEDULE: HandlerInvokeType
RUN_INTERVAL: HandlerInvokeType
WEBHOOK: HandlerInvokeType
DISPATCH_MESSAGE_TYPE_NONE: DispatchMessageType
DISPATCH_MESSAGE_INVOKE: DispatchMessageType
DISPATCH_MESSAGE_WORK_COMPLETED: DispatchMessageType

class RegisterHandlerRequest(_message.Message):
    __slots__ = ('dispatch_id', 'handler_name', 'timeout_ms', 'options')
    DISPATCH_ID_FIELD_NUMBER: _ClassVar[int]
    HANDLER_NAME_FIELD_NUMBER: _ClassVar[int]
    TIMEOUT_MS_FIELD_NUMBER: _ClassVar[int]
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    dispatch_id: str
    handler_name: str
    timeout_ms: int
    options: _containers.RepeatedCompositeFieldContainer[HandlerOption]

    def __init__(self, dispatch_id: _Optional[str]=..., handler_name: _Optional[str]=..., timeout_ms: _Optional[int]=..., options: _Optional[_Iterable[_Union[HandlerOption, _Mapping]]]=...) -> None:
        ...

class HandlerInvokeOption(_message.Message):
    __slots__ = ('type', 'value')
    TYPE_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    type: HandlerInvokeType
    value: str

    def __init__(self, type: _Optional[_Union[HandlerInvokeType, str]]=..., value: _Optional[str]=...) -> None:
        ...

class HandlerOption(_message.Message):
    __slots__ = ('invoke',)
    INVOKE_FIELD_NUMBER: _ClassVar[int]
    invoke: HandlerInvokeOption

    def __init__(self, invoke: _Optional[_Union[HandlerInvokeOption, _Mapping]]=...) -> None:
        ...

class RegisterHandlerResponse(_message.Message):
    __slots__ = ('error', 'id')
    ERROR_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    error: _common_pb2.Error
    id: str

    def __init__(self, error: _Optional[_Union[_common_pb2.Error, _Mapping]]=..., id: _Optional[str]=...) -> None:
        ...

class UnregisterHandlerRequest(_message.Message):
    __slots__ = ('id',)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str

    def __init__(self, id: _Optional[str]=...) -> None:
        ...

class UnregisterHandlerResponse(_message.Message):
    __slots__ = ('error',)
    ERROR_FIELD_NUMBER: _ClassVar[int]
    error: _common_pb2.Error

    def __init__(self, error: _Optional[_Union[_common_pb2.Error, _Mapping]]=...) -> None:
        ...

class ListHandlersRequest(_message.Message):
    __slots__ = ()

    def __init__(self) -> None:
        ...

class HandlerInfo(_message.Message):
    __slots__ = ('name', 'options', 'dispatch_id', 'id', 'last_invoked_client_timestamp', 'is_active')
    NAME_FIELD_NUMBER: _ClassVar[int]
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    DISPATCH_ID_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    LAST_INVOKED_CLIENT_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    IS_ACTIVE_FIELD_NUMBER: _ClassVar[int]
    name: str
    options: _containers.RepeatedCompositeFieldContainer[HandlerOption]
    dispatch_id: str
    id: str
    last_invoked_client_timestamp: _timestamp_pb2.Timestamp
    is_active: bool

    def __init__(self, name: _Optional[str]=..., options: _Optional[_Iterable[_Union[HandlerOption, _Mapping]]]=..., dispatch_id: _Optional[str]=..., id: _Optional[str]=..., last_invoked_client_timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]]=..., is_active: bool=...) -> None:
        ...

class ListHandlersResponse(_message.Message):
    __slots__ = ('error', 'handlers')
    ERROR_FIELD_NUMBER: _ClassVar[int]
    HANDLERS_FIELD_NUMBER: _ClassVar[int]
    error: _common_pb2.Error
    handlers: _containers.RepeatedCompositeFieldContainer[HandlerInfo]

    def __init__(self, error: _Optional[_Union[_common_pb2.Error, _Mapping]]=..., handlers: _Optional[_Iterable[_Union[HandlerInfo, _Mapping]]]=...) -> None:
        ...

class DispatchRequest(_message.Message):
    __slots__ = ('dispatch_id', 'client_version')
    DISPATCH_ID_FIELD_NUMBER: _ClassVar[int]
    CLIENT_VERSION_FIELD_NUMBER: _ClassVar[int]
    dispatch_id: str
    client_version: str

    def __init__(self, dispatch_id: _Optional[str]=..., client_version: _Optional[str]=...) -> None:
        ...

class DispatchMessage(_message.Message):
    __slots__ = ('type', 'invoke')
    TYPE_FIELD_NUMBER: _ClassVar[int]
    INVOKE_FIELD_NUMBER: _ClassVar[int]
    type: DispatchMessageType
    invoke: DispatchHandlerInvoke

    def __init__(self, type: _Optional[_Union[DispatchMessageType, str]]=..., invoke: _Optional[_Union[DispatchHandlerInvoke, _Mapping]]=...) -> None:
        ...

class DispatchHandlerInvoke(_message.Message):
    __slots__ = ('invocation_id', 'dispatch_id', 'handler_id', 'handler_name', 'timeout_ms', 'reason', 'args')

    class ArgsEntry(_message.Message):
        __slots__ = ('key', 'value')
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str

        def __init__(self, key: _Optional[str]=..., value: _Optional[str]=...) -> None:
            ...
    INVOCATION_ID_FIELD_NUMBER: _ClassVar[int]
    DISPATCH_ID_FIELD_NUMBER: _ClassVar[int]
    HANDLER_ID_FIELD_NUMBER: _ClassVar[int]
    HANDLER_NAME_FIELD_NUMBER: _ClassVar[int]
    TIMEOUT_MS_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    ARGS_FIELD_NUMBER: _ClassVar[int]
    invocation_id: str
    dispatch_id: str
    handler_id: str
    handler_name: str
    timeout_ms: int
    reason: HandlerInvokeType
    args: _containers.ScalarMap[str, str]

    def __init__(self, invocation_id: _Optional[str]=..., dispatch_id: _Optional[str]=..., handler_id: _Optional[str]=..., handler_name: _Optional[str]=..., timeout_ms: _Optional[int]=..., reason: _Optional[_Union[HandlerInvokeType, str]]=..., args: _Optional[_Mapping[str, str]]=...) -> None:
        ...

class Log(_message.Message):
    __slots__ = ('level', 'timestamp', 'message')
    LEVEL_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    level: str
    timestamp: _timestamp_pb2.Timestamp
    message: str

    def __init__(self, level: _Optional[str]=..., timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]]=..., message: _Optional[str]=...) -> None:
        ...

class ReportInvocationRequest(_message.Message):
    __slots__ = ('handler_invoke', 'start_client_timestamp', 'duration_ms', 'result', 'error', 'logs')
    HANDLER_INVOKE_FIELD_NUMBER: _ClassVar[int]
    START_CLIENT_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    DURATION_MS_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    LOGS_FIELD_NUMBER: _ClassVar[int]
    handler_invoke: DispatchHandlerInvoke
    start_client_timestamp: _timestamp_pb2.Timestamp
    duration_ms: int
    result: InvokeResult
    error: _common_pb2.Error
    logs: _containers.RepeatedCompositeFieldContainer[Log]

    def __init__(self, handler_invoke: _Optional[_Union[DispatchHandlerInvoke, _Mapping]]=..., start_client_timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]]=..., duration_ms: _Optional[int]=..., result: _Optional[_Union[InvokeResult, _Mapping]]=..., error: _Optional[_Union[_common_pb2.Error, _Mapping]]=..., logs: _Optional[_Iterable[_Union[Log, _Mapping]]]=...) -> None:
        ...

class ReportInvocationResponse(_message.Message):
    __slots__ = ('error',)
    ERROR_FIELD_NUMBER: _ClassVar[int]
    error: _common_pb2.Error

    def __init__(self, error: _Optional[_Union[_common_pb2.Error, _Mapping]]=...) -> None:
        ...

class InvokeResult(_message.Message):
    __slots__ = ('value',)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: str

    def __init__(self, value: _Optional[str]=...) -> None:
        ...

class GetHandlerHistoryRequest(_message.Message):
    __slots__ = ('handler_name', 'start_time', 'end_time', 'include_logs', 'tail')
    HANDLER_NAME_FIELD_NUMBER: _ClassVar[int]
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    END_TIME_FIELD_NUMBER: _ClassVar[int]
    INCLUDE_LOGS_FIELD_NUMBER: _ClassVar[int]
    TAIL_FIELD_NUMBER: _ClassVar[int]
    handler_name: str
    start_time: _timestamp_pb2.Timestamp
    end_time: _timestamp_pb2.Timestamp
    include_logs: bool
    tail: int

    def __init__(self, handler_name: _Optional[str]=..., start_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]]=..., end_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]]=..., include_logs: bool=..., tail: _Optional[int]=...) -> None:
        ...

class HandlerExecution(_message.Message):
    __slots__ = ('handler_name', 'handler_id', 'invocation_id', 'dispatch_id', 'publish_server_timestamp', 'receive_server_timestamp', 'start_client_timestamp', 'duration_ms', 'error', 'logs')
    HANDLER_NAME_FIELD_NUMBER: _ClassVar[int]
    HANDLER_ID_FIELD_NUMBER: _ClassVar[int]
    INVOCATION_ID_FIELD_NUMBER: _ClassVar[int]
    DISPATCH_ID_FIELD_NUMBER: _ClassVar[int]
    PUBLISH_SERVER_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    RECEIVE_SERVER_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    START_CLIENT_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    DURATION_MS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    LOGS_FIELD_NUMBER: _ClassVar[int]
    handler_name: str
    handler_id: str
    invocation_id: str
    dispatch_id: str
    publish_server_timestamp: _timestamp_pb2.Timestamp
    receive_server_timestamp: _timestamp_pb2.Timestamp
    start_client_timestamp: _timestamp_pb2.Timestamp
    duration_ms: int
    error: _common_pb2.Error
    logs: _containers.RepeatedCompositeFieldContainer[Log]

    def __init__(self, handler_name: _Optional[str]=..., handler_id: _Optional[str]=..., invocation_id: _Optional[str]=..., dispatch_id: _Optional[str]=..., publish_server_timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]]=..., receive_server_timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]]=..., start_client_timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]]=..., duration_ms: _Optional[int]=..., error: _Optional[_Union[_common_pb2.Error, _Mapping]]=..., logs: _Optional[_Iterable[_Union[Log, _Mapping]]]=...) -> None:
        ...

class GetHandlerHistoryResponse(_message.Message):
    __slots__ = ('error', 'history')
    ERROR_FIELD_NUMBER: _ClassVar[int]
    HISTORY_FIELD_NUMBER: _ClassVar[int]
    error: _common_pb2.Error
    history: _containers.RepeatedCompositeFieldContainer[HandlerExecution]

    def __init__(self, error: _Optional[_Union[_common_pb2.Error, _Mapping]]=..., history: _Optional[_Iterable[_Union[HandlerExecution, _Mapping]]]=...) -> None:
        ...


"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(_runtime_version.Domain.PUBLIC, 5, 28, 1, '', 'cortex-api.proto')
_sym_db = _symbol_database.Default()
DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10cortex-api.proto\x12\x0bcortex.axon"O\n\x0bCallRequest\x12\x0e\n\x06method\x18\x02 \x01(\t\x12\x0c\n\x04path\x18\x03 \x01(\t\x12\x14\n\x0ccontent_type\x18\x04 \x01(\t\x12\x0c\n\x04body\x18\x05 \x01(\t"\xaa\x01\n\x0cCallResponse\x12\x13\n\x0bstatus_code\x18\x01 \x01(\x05\x12\x0e\n\x06status\x18\x02 \x01(\t\x127\n\x07headers\x18\x03 \x03(\x0b2&.cortex.axon.CallResponse.HeadersEntry\x12\x0c\n\x04body\x18\x04 \x01(\t\x1a.\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x028\x012H\n\tCortexApi\x12;\n\x04Call\x12\x18.cortex.axon.CallRequest\x1a\x19.cortex.axon.CallResponseB\x1cZ\x1agithub.com/cortexapps/axonb\x06proto3')
_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'cortex_api_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    _globals['DESCRIPTOR']._loaded_options = None
    _globals['DESCRIPTOR']._serialized_options = b'Z\x1agithub.com/cortexapps/axon'
    _globals['_CALLRESPONSE_HEADERSENTRY']._loaded_options = None
    _globals['_CALLRESPONSE_HEADERSENTRY']._serialized_options = b'8\x01'
    _globals['_CALLREQUEST']._serialized_start = 33
    _globals['_CALLREQUEST']._serialized_end = 112
    _globals['_CALLRESPONSE']._serialized_start = 115
    _globals['_CALLRESPONSE']._serialized_end = 285
    _globals['_CALLRESPONSE_HEADERSENTRY']._serialized_start = 239
    _globals['_CALLRESPONSE_HEADERSENTRY']._serialized_end = 285
    _globals['_CORTEXAPI']._serialized_start = 287
    _globals['_CORTEXAPI']._serialized_end = 359


from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional
DESCRIPTOR: _descriptor.FileDescriptor

class Error(_message.Message):
    __slots__ = ('code', 'message')
    CODE_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    code: str
    message: str

    def __init__(self, code: _Optional[str]=..., message: _Optional[str]=...) -> None:
        ...


import json
import queue
import traceback
from datetime import datetime
from time import sleep
from typing import Any, Optional
from urllib.parse import quote_plus, urlencode

import grpc

from generated import (
    common_pb2,
    cortex_api_pb2,
    cortex_api_pb2_grpc,
    cortex_axon_agent_pb2,
    cortex_axon_agent_pb2_grpc,
)

from .agentversion import AGENT_VERSION
from .handler import (
    CortexAnnotation,
    find_annotated_methods,
)


# TODO Add docstrings
class CortexRequestException(IOError):
    """There was an ambiguous exception that occurred while handling your
    request.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class CortexHTTPError(CortexRequestException):
    """An HTTP error occurred."""


class CortexResponse:
    """
    A class to represent the response from a Cortex API call.

    Attributes:
    -----------
    status_code : int
        The HTTP status code of the response.
    text : str
        The body of the response. In plaintext
    headers : Dict[str, str]
        The headers of the response.
    """

    def __init__(self, status_code: int, text: str, headers: dict[str, str]):
        """
        Constructs all the necessary attributes for the CortexResponse object.

        Parameters:
        -----------
        status_code : int
            The HTTP status code of the response.
        body : str
            The body of the response.
        headers : Dict[str, str]
            The headers of the response.
        """
        self.status_code = status_code
        self.text = text
        self.headers = headers

    def json(self) -> Any:
        """
        Returns the JSON representation of the response body.

        Returns:
        --------
        Any
            The JSON representation of the response body.
        """
        return json.loads(self.text)

    def raise_for_status(self):
        """Raises :class:`HTTPError`, if one occurred."""

        http_error_msg = ''
        if 400 <= self.status_code < 500:
            http_error_msg = f'{self.status_code} Client Error: {self.text}'

        elif 500 <= self.status_code < 600:
            http_error_msg = f'{self.status_code} Server Error: {self.text}'

        if http_error_msg:
            raise CortexHTTPError(http_error_msg)

    @property
    def ok(self):
        """Returns True if :attr:`status_code` is less than 400, False if not.

        This attribute checks if the status code of the response is between
        400 and 600 to see if there was a client error or a server error. If
        the status code is between 200 and 400, this will return True. This
        is **not** a check to see if the response code is ``200 OK``.
        """
        try:
            self.raise_for_status()
        except CortexHTTPError:
            return False
        return True


def _client_message_iterator(initial_message: cortex_axon_agent_pb2.DispatchRequest):
    blocking_queue = queue.Queue()
    blocking_queue.put(initial_message)
    while True:
        message = blocking_queue.get()
        yield message


class AxonClient:
    def __init__(
        self,
        agent_host: str = "localhost",
        agent_port: int = 50051,
        cortex_host: str = None,
        cortex_port: int = None,
        handlers: Optional[list[CortexAnnotation]] = None,
        scope=None,
    ):
        self.id = str(datetime.now().timestamp())
        self.agent_hostport = f"{agent_host}:{agent_port}"

        cortex_api_host = cortex_host or agent_host
        cortex_api_port = cortex_port or agent_port
        self.cortex_hostport = f"{cortex_api_host}:{cortex_api_port}"

        self.cortex_channel = grpc.insecure_channel(self.cortex_hostport)
        self.cortex_stub = cortex_api_pb2_grpc.CortexApiStub(
            self.cortex_channel)

        self.agent_channel = None
        self.agent_stub = None

        if not scope and not handlers:
            raise ValueError("One of scope (globals()) or handlers required")

        self.handlers = handlers or find_annotated_methods(scope)

    def _get_stub(self):
        if not self.agent_stub:
            self.agent_channel = grpc.insecure_channel(self.agent_hostport)
            self.agent_stub = cortex_axon_agent_pb2_grpc.AxonAgentStub(
                self.agent_channel
            )
        return self.agent_stub

    def _clear_stub(self):
        if self.agent_stub:
            self.agent_channel.close()
            self.agent_stub = None

    def _register_handlers(self):
        for h in self.handlers:
            options = h.get_handler_options()

            print(f"Registering handler: {h.name}")
            self.agent_stub.RegisterHandler(
                cortex_axon_agent_pb2.RegisterHandlerRequest(
                    dispatch_id=self.id,
                    handler_name=h.name,
                    timeout_ms=h.timeout,
                    options=options,
                )
            )

    def _find_handler(self, name):
        for h in self.handlers:
            if h.name == name:
                return h
        return None

    def _handler_context(self, args):
        return HandlerContext(self.cortex_stub, args)

    def run(self):
        registered = False
        while True:
            stub = self._get_stub()
            try:
                if not registered:
                    self._register_handlers()
                    registered = True

                initial_message = cortex_axon_agent_pb2.DispatchRequest(
                    dispatch_id=self.id,
                    client_version=AGENT_VERSION
                )
                for response in stub.Dispatch(
                    _client_message_iterator(initial_message)
                ):

                    if response.type == cortex_axon_agent_pb2.DISPATCH_MESSAGE_INVOKE:
                        invoke = response.invoke
                        args = invoke.args
                        handler_to_invoke = self._find_handler(
                            invoke.handler_name)
                        if not handler_to_invoke:
                            print(f"Unknown handler: {response.handler_name}")
                            continue
                        print(
                            f"Dispatch handler START {invoke.handler_name} (id={invoke.invocation_id}) reason={invoke.reason}"
                        )
                        handler_err = None
                        handler_result = None
                        result = None
                        now = datetime.now()
                        ctx = self._handler_context(args)
                        duration_ms = 0

                        try:
                            result = handler_to_invoke(ctx)
                            duration_ms = int(
                                (datetime.now() - now).total_seconds() * 1000
                            )
                            if result:
                                handler_result = cortex_axon_agent_pb2.InvokeResult(
                                    value=str(result)
                                )
                            print(f"Dispatch handler SUCCESS {invoke.handler_name} (id={invoke.invocation_id}) in {round((datetime.now() - now).total_seconds() * 1000, 1)}ms. Returned result={result is not None}")
                        except Exception as e:
                            print(f"Dispatch handler ERROR {invoke.handler_name} (id={invoke.invocation_id}) in {round((datetime.now() - now).total_seconds() * 1000, 1)}ms. Error={e}")
                            traceback.print_exc()
                            handler_err = common_pb2.Error(
                                code="unexpected", message="Error calling handler: " + str(e)
                            )

                        invoke_info = cortex_axon_agent_pb2.ReportInvocationRequest(
                            handler_invoke=invoke,
                            start_client_timestamp=now,
                            duration_ms=duration_ms,
                            result=handler_result,
                            error=handler_err,
                            logs=ctx.logs,
                        )
                        report_response = self.agent_stub.ReportInvocation(
                            invoke_info)

                        if report_response.error and report_response.error.code:
                            print(
                                f"Error reporting invocation: {report_response.error}"
                            )
                    elif (
                        response.type == cortex_axon_agent_pb2.DISPATCH_MESSAGE_WORK_COMPLETED
                    ):
                        print("Work completed, shutting down...")
                        return
                    else:
                        print(f"Unknown message type: {response.type}")

                print("Response stream disconnected... retrying in 5 seconds...")
                sleep(5)
            except Exception as e:
                print("Error calling axon server, retrying in 5 seconds...")
                print(e)
                traceback.print_exc()
                self._clear_stub()
                registered = False
                sleep(5)
                continue


class HandlerContext:

    api: cortex_api_pb2_grpc.CortexApiStub
    logs: list[cortex_axon_agent_pb2.Log]
    args: dict
    obj: dict[str, Any]

    def __init__(self, api: cortex_api_pb2_grpc.CortexApiStub, args: dict):
        self.api = api
        self.args = args
        self.logs = []
        self.obj = {}

    def log(self, message: str, level="INFO"):
        """
        Log a message with a given level.
        :param message:
        :param level:
        """
        self.logs.append(
            cortex_axon_agent_pb2.Log(
                timestamp=datetime.now(), level=level, message=message
            )
        )
        print(f"{level}: {message}")

    def cortex_api_call(
        self, path: str, method: str, body: str = None, content_type: str = "application/json", params: dict[str, str] = None
    ) -> CortexResponse:
        if params:
            encoded_params = {k: quote_plus(v) for k, v in params.items()}
            query_string = urlencode(encoded_params)
            path = f"{path}?{query_string}"

        request = cortex_api_pb2.CallRequest(
            method=method,
            path=path,
            body=body,
            content_type=content_type,
        )
        response = self.api.Call(request)
        return CortexResponse(response.status_code, response.body, response.headers)
